纯python比numpy更快的数据类型转换

aer*_*man 3 python performance numpy vectorization

请原谅我的无知。

如果numpy提供向量化运算以加快计算速度,那么对于数据类型转换而言,纯python的速度几乎快8倍吗?

例如

a = np.random.randint(0,500,100).astype(str)
b = np.random.randint(0,500,100).astype(str)
c = np.random.randint(0,500,100).astype(str)

def A(a,b,c):
    for i,j,k in zip(a,b,c):
        d,e,f = int(i), int(j), int(k)
        r = d+e-f
    return 

def B(a,b,c):
    for i,j,k in zip(a,b,c):
        d,e,f  = np.array([i,j,k]).astype(int)
        r = d+e-f
    return 
Run Code Online (Sandbox Code Playgroud)

然后,

%%timeit 
A(a,b,c)
Run Code Online (Sandbox Code Playgroud)

每个回路249 µs±3.13 µs(平均±标准偏差,共运行7次,每个回路1000个)

%%timeit
B(a,b,c)
Run Code Online (Sandbox Code Playgroud)

每个循环1.87 ms±4.08 µs(平均±标准偏差,共运行7次,每个循环1000个)

谢谢你,爱丽儿

use*_*203 8

是的,NumPy 确实提供矢量化运算,这些运算使计算速度比原始Python代码更快。但是,您没有使用它们。

NumPy旨在跨整个数据集执行操作,跨数据集块执行的重复操作不多。后者导致迭代在Python级别完成,这将增加运行时间。

您的主要问题是,您正在使用的唯一“向量化”操作是astype,但是您一次将其应用于三个元素,并且仍然像幼稚的Python解决方案一样循环。再加上在循环的每次迭代中创建numpy数组都会产生额外开销的事实,因此使用numpy进行尝试的速度变慢也就不足为奇了。

在小型数据集上,Python 可能会更快,因为NumPy具有创建数组,将对象与低级库之间来回传递等开销。让我们来看一下一次在三个元素上使用的强制转换操作:

%timeit np.array(['1', '2', '3']).astype(int)
5.25 µs ± 89.3 ns per loop (mean ± std. dev. of 7 runs, 100000 loops each)

%timeit np.array(['1', '2', '3'])
1.62 µs ± 42.9 ns per loop (mean ± std. dev. of 7 runs, 1000000 loops each)
Run Code Online (Sandbox Code Playgroud)

超过四分之一的运行时间仅来自于分配数组!将此与您的纯Python版本进行比较:

%timeit a, b, c = int('1'), int('2'), int('3')
659 ns ± 50.7 ns per loop (mean ± std. dev. of 7 runs, 1000000 loops each)
Run Code Online (Sandbox Code Playgroud)

因此,如果仅对这种大小的块进行操作,Python将击败NumPy。


但是,除了三个元素之外,您还有更多的元素,因此可以使用NumPy 大大加快代码的速度,但是您需要改变对解决问题的态度。与其关注于如何将运算应用于单个标量,不如考虑如何将运算应用于数组。


为了解决这个问题,总体思路是:

  • 创建一个包含所有值的数组
  • int一次astype调用即可将整个数组转换为。
  • 提前进行逐元素运算,以将所需的算术应用于数组。

最终看起来像这样:

def vectorized(a, b, c):
    u = np.array([a, b, c]).astype(int)
    return u[0] + u[1] - u[2]
Run Code Online (Sandbox Code Playgroud)

一旦比较了正确使用NumPy的两种方法,您将开始看到性能大幅提高。

def python_loop(a, b, c):
    out = []
    for i,j,k in zip(a,b,c):
        d,e,f = int(i), int(j), int(k)
        out.append(d+e-f)
    return out

a, b, c = np.random.randint(0, 500, (3, 100_000)).astype(str)

In [255]: %timeit vectorized(a, b, c)
181 ms ± 6.07 ms per loop (mean ± std. dev. of 7 runs, 10 loops each)

In [256]: %timeit python_loop(a, b, c)
206 ms ± 7.97 ms per loop (mean ± std. dev. of 7 runs, 10 loops each)

>>> np.array_equal(python_loop(a, b, c), vectorized(a, b, c))
True
Run Code Online (Sandbox Code Playgroud)

从字符串转换为整数并不意味着NumPy会比纯Python快得多,从时间上可以看出,两者相当接近。但是,通过应用矢量化方法,比较至少要公平得多。